
# circleLimit.py
# Andrew Davison, ad@coe.psu.ac.th, Sept. 2024
'''
  Generates circleLimit.png

  https://www.josleys.com/show_gallery.php?galid=325

  Circle Limit I and Circle Limit IV are subgroups
  of the group [6,4] (in Coxeter notation), which is generated 
  by reflections across the sides of a 30-45-90 degree 
  hyperbolic triangle.

  Similarly, Circle Limit I and Circle Limit IV are
  subgroups of the group [8,3], which is generated by 
  reflections across the sides of a 22.5-60-90 degree 
  hyperbolic triangle.

  "How Did Escher Do It?", Bill Casselman,
  http://www.ams.org/publicoutreach/feature-column/fcarc-circle-limit

  https://en.wikipedia.org/wiki/Circle_Limit_III
  https://eschermath.org/wiki/Circle_Limit_Exploration.html
  http://www.ams.org/publicoutreach/feature-column/fcarc-circle-limit

  https://www.d.umn.edu/~ddunham/
'''

import math

from geom import *
from pgeom import *


def mirrorEdge(tri, edgeNo):
  edge = tri.getEdge(edgeNo)
  return Transform.mirror(edge.startPt(), edge.endPt())


def getMidpoint(tri):
  midLine0 = Line.points(tri.getVertex(0), 
                         tri.getEdge(1).midpoint(), segment=True)
  midLine1 = Line.points(tri.getVertex(1), 
                         tri.getEdge(2).midpoint(), segment=True)
  midPt = midLine0.intersectPts(midLine1)[0]
  # print(midPt)
  return midPt


def getMirrors(tri, midPts):
  mmPts = []
  mTris = []
  for i in range(3):
    mTri = tri.transform(mirrorEdge(tri, i))
    midPt = getMidpoint(mTri)
    eDist = math.hypot(*midPt)
    if (midPt not in midPts) and (eDist < 0.95):
      mmPts.append(midPt)
      angle = int(math.degrees(midPt.theta))
      mTris.append(mTri)
      d.draw(mTri)
      # if angle%2 == 0:
        # d.draw(mTri, fill='pink')
      # else:
        # d.draw(mTri, fill='tan')
  return mmPts, mTris


d = createDrawing()
d.draw(ECircle(0, 0, 1), stroke='silver') # disc boundary

# p = 8
# q = 3
print("Some (p,q)'s: (3,7), (4,5), (4,6), (6,4), (8,3), (10,3)...")
print("where (p-2) * (q-2) > 4")
p, q = map(int, input("p q? ").split())

r = polyRadius(p, q)
# print("r", r)
p0 = Point(0,0)
p1 = Point.h_polar(r, deg=0)
p2 = Point.h_polar(r, deg=360/p)

tri1 = Triangle(p0,p1,p2)
d.draw(tri1)

midPts = [getMidpoint(tri1)]
tris = [tri1]

while len(tris) != 0:
  tri = tris.pop(0)
  mmPts, mTris = getMirrors(tri, midPts)
  midPts.extend(mmPts)
  tris.extend(mTris)

print("No. midpts generated:", len(midPts))

saveDrawing(d, f"circle_{p}_{q}.png")

