
# geom.py
# Andrew Davison, ad@coe.psu.ac.th, Sept. 2024
'''
  Euclidean geometry classes:
     ESegment, ECircle, EArc
     intersection functions,
     other utilities

  Derived from
    https://github.com/cduck/hyperbolic/tree/master


  and 
   Eline, EVertical and utilities
   From: Section 16.7 (was 16.8); hyperbolic_tessellation
    "A Programmer’s Introduction to Mathematics"
    Jeremy Kun, 2021
    https://pimbook.org/
    https://github.com/pim-book/programmers-introduction-to-mathematics

  and 
    "Geometry"
    David A. Brannan, et al,
    Cambridge University Press; 2nd edition

  and
    https://en.wikipedia.org/wiki/Inversive_geometry
    https://mathworld.wolfram.com/Inversion.html

'''

import math
import numpy.linalg
import drawsvg

import matplotlib
import matplotlib.pyplot as plt 
import matplotlib.image as img 

EPSILON = 1e-9

WIN_WIDTH = 2.1
RENDER_SIZE = 400

maxVal = 100

# ----------------------------------------------

class ELine:
  def __init__(self, x, y, slope, lineSize=maxVal):
    self.lineSize = lineSize
    self.x = x
    self.y = y
    self.slope = slope


  def __str__(self):
    return f"ELine(({self.x:.3f},{self.y:.3f}), slope={self.slope:.3f})"

  def __repr__(self):
    return str(self)


  def __eq__(self, other):
    if not isinstance(other, ELine):
      return False
    if isinstance(other, EVertical):
      return other.__eq__(self)
    return (
      nearZero(self.slope - other.slope) and \
      self.contains(other.x, other.y)
    )

  def __ne__(self, other):
    return not self.__eq__(other)


  def yValue(self, x):
    # Compute the y value of the x pt on this line
    return self.slope*(x - self.x) + self.y

  def getCoords(self, n=20):
    xStep = 2*maxVal/n
    pts = []
    x = -maxVal
    for i in range(n):
      pts.append( (x, self.yValue(x)))
      x += xStep
    return pts


  def toESegment():
    return ESegment(-maxVal, self.yValue(-maxVal), 
                    maxVal, self.yValue(maxVal))

  @staticmethod
  def through(x1, y1, x2, y2):
    # Return a ELine through the two given points
    if nearZero(x1 - x2):
      return EVertical.at(x1)
    return ELine(x1, y1, (y2 - y1)/(x2 - x1))


  def intersects(self, line):
    # Compute the intersection of two lines
    if isinstance(line, EVertical):
      return line.intersects(self)

    if self == line:
      raise ValueError("Can't intersect two identical lines; "
               "solution is not a single pt.")

    if nearZero(self.slope) and nearZero(line.slope):
      raise ValueError("Can't intersect two horizontal lines.")

    x1, y1, slope1 = self.x, self.y, self.slope
    x2, y2, slope2 = line.x, line.y, line.slope

    xInter = ((slope1*x1 - slope2*x2 + y2 - y1)
             /(slope1 - slope2))
    yInter = y1 + slope1*(xInter - x1)
    return xInter, yInter


  def angleTo(self, line):
    # Compute the angle between two lines
    if isinstance(line, EVertical):
      return 90 - math.degrees(math.atan(self.slope))

    if self == line:
      return 0
    if nearZero(self.slope) and nearZero(line.slope):
      return 0

    # https://byjus.com/maths/angle-between-two-lines/
    m1 = self.slope
    m2 = line.slope
    return math.degrees(math.atan2((1 + m1*m2), (m2-m1)))


  def contains(self, x, y):
    return nearZero(self.yValue(x) - y)

  def reflect(self, x, y):
    # https://stackoverflow.com/questions/3306838/algorithm-for-reflecting-a-point-across-a-line
    m = self.slope
    c = self.y - m*self.x
    d = (x + (y - c)*m)/(1 + m*m)
    xRef = 2*d - x
    yRef = 2*d*m - y + 2*c
    return xRef, yRef


  def invert(self, ic):
    # invert the line to create a circle or a line
    pts1 = self.getCoords(10)
    iPts = []
    for i, pt in enumerate(pts1):
      ipt = invertPt(*pt, ic.cx, ic.cy, ic.r)
      if not ipt == None:
        iPts.append(ipt)
    if orientation(*iPts[0], *iPts[1], *iPts[2]) == 'collinear':
      return ELine.through(*iPts[0], *iPts[1])
    else:
      return ECircle.points(*iPts[0], *iPts[1], *iPts[2])


  def to_drawables(self, **kwargs):
    width = self.lineSize
    x1 = -width
    y1 = self.yValue(x1)
    x2 = width
    y2 = self.yValue(x2)
    # print(x1, y1, x2, y2)
    return (drawsvg.Line(x1, y1, x2, y2, **kwargs),)


  def draw_to_path(self, path, include_m=True, include_l=False):
    width = self.lineSize
    x1 = -width
    y1 = self.yValue(x1)
    x2 = width
    y2 = self.yValue(x2)
    if include_l:
      path.L(x1, y1)
    elif include_m:
      path.M(x1, y1)
    path.L(x2, y2)


# ----------------------------------------------

class EVertical(ELine):
  @staticmethod
  def at(x):
    return EVertical(x, 0, "vertical")

  def yValue(self, x):
    raise TypeError("EVertical does not support a y-value")

  def reflect(self, x, y):
    # Reflect a pt across this line
    return 2*self.x - x, y

  def __eq__(self, other):
    return (isinstance(other, EVertical) and \
            nearZero(self.x - other.x) )

  def __str__(self):
    return f"Vertical(x={self.x:.3f})"


  def getCoords(self, n=20):
    yStep = 2*maxVal/n
    pts = []
    y = -maxVal
    for i in range(n):
      pts.append( (self.x, y))
      y += yStep
    return pts


  def intersects(self, line):
    # pt of intersection of this line with another
    if isinstance(line, EVertical):
      raise ValueError("Can't intersect two vertical lines; "
               "solution is either empty or not a pt.")
    return self.x, line.yValue(self.x)


  def to_drawables(self, **kwargs):
    x = self.x
    height = self.lineSize
    return (drawsvg.Line(x, -height, x, height, **kwargs),)

  def draw_to_path(self, path, include_m=True, include_l=False):
    x = self.x
    height = self.lineSize
    if include_l:
      path.L(x, -height)
    elif include_m:
      path.M(x, -height)
    path.L(x, height)


# ----------------------------------------------

class ESegment:

  def __init__(self, x1, y1, x2, y2):
    self.x1 = x1
    self.y1 = y1
    self.x2 = x2
    self.y2 = y2

  def __str__(self):
    return f'{type(self).__name__}(({self.x1:.3f}, {self.y1:.3f}) -- ({self.x2:.3f}, {self.y2:.3f}))'

  def __repr__(self):
    return str(self)


  def reverse(self):
    self.x1, self.y1, self.x2, self.y2 =  \
              self.x2, self.y2, self.x1, self.y1

  def reversed(self):
    return ESegment(self.x2, self.y2, self.x1, self.y1)

  def atan2(self):
    return math.atan2(self.y2-self.y1, self.x2-self.x1)

  def length(self):
    return math.dist((x1,y1), (x2,y2))
    # ((self.x2 - self.x1)**2 + (self.y2 - self.y1)**2) ** 0.5


  def parallelTo(self, seg2):
    return nearZero(math.pi/2 -
        (self.atan2() - seg2.atan2() + math.pi/2) % math.pi)

  def antiparallelTo(self, seg2):
    # https://en.wikipedia.org/wiki/Antiparallel_lines
    return nearZero(math.pi -
        (self.atan2() - seg2.atan2()) % (math.pi*2))



  def startPt(self):
    return self.x1, self.y1

  def endPt(self):
    return self.x2, self.y2

  def midpoint(self):
    return (self.x1 + self.x2)/2, (self.y1 + self.y2)/2

  def toELine():
    return ELine.through(self.x1, self.y1, self.x2, self.y2)


  def trimmed(self, x1, y1, x2, y2, **kwargs):
    # new segment with these endpoints
    return ESegment(x1, y1, x2, y2)


  def perpendicular(self, x, y, length=1):
    '''Return a segment through (x,y) rotated 
       90 deg clockwise from self.'''
    x1, y1, x2, y2 = self.x1, self.y1, self.x2, self.y2
    swap = nearZero(x-x1) and nearZero(y-y1)
    if swap:
      x1, y1, x2, y2 = x2, y2, x1, y1

    lineAngle = math.atan2(y2-y1, x2-x1)
    d = math.hypot(x-x1, y-y1)
    ang = math.atan2(y-y1, x-x1)
    d_along = d * math.cos(ang - lineAngle)
    px1 = d_along * math.cos(lineAngle)
    py1 = d_along * math.sin(lineAngle)
    d2 = math.hypot(d_along, length)
    ang2 = math.atan2(length, d_along) + lineAngle
    px2 = d2 * math.cos(ang2)
    py2 = d2 * math.sin(ang2)
    if swap:
      px2 = 2*px1 - px2
      py2 = 2*py1 - py2
    return ESegment(x1+px1, y1+py1, x1+px2, y1+py2)


  def onSegment(self, x, y):
    ''' Tests if the point is between the segment 
        endpoints, assuming it is
        already on the infinite-length line. '''
    # Dot product of vectors start->(x,y) and start->end
    k1 = ((x - self.x1) * (self.x2 - self.x1)
        + (y - self.y1) * (self.y2 - self.y1))

    # Dot product of vectors end->(x,y) and end->start
    k2 = ((x - self.x2) * (self.x1 - self.x2)
        + (y - self.y2) * (self.y1 - self.y2))
    return k1 >= 0 and k2 >= 0


  @staticmethod
  def radicalAxis(circ1, circ2):
    # https://en.wikipedia.org/wiki/Radical_axis
    '''Return the radical axis for two circles (or points).
       If one input is a line, the output is this line. '''
    if not isinstance(circ1, (ECircle, ESegment)):
      circ1 = ECircle(*circ1, 0)

    if not isinstance(circ2, (ECircle, ESegment)):
      circ2 = ECircle(*circ2, 0)

    if isinstance(circ1, ESegment) and isinstance(circ2, ESegment):
      raise ValueError('No radical axis exists for two lines')
    elif isinstance(circ1, ESegment):
      # TODO: Possibly correct direction to be consistent
      return circ1
    elif isinstance(circ2, ESegment):
      # TODO: Possibly correct direction to be consistent
      return circ2
    else:
      cx1, cy1, r1 = circ1.cx, circ1.cy, circ1.r
      cx2, cy2, r2 = circ2.cx, circ2.cy, circ2.r
      d = math.hypot(cx2-cx1, cy2-cy1)
      if nearZero(d):
        raise ValueError('Circles are concentric')
      d1 = (d**2 + r1**2 - r2**2) / (2*d)
      ratio = d1 / d
      mx = cx1 + ratio * (cx2 - cx1)
      my = cy1 + ratio * (cy2 - cy1)
      return ESegment(cx1, cy1, cx2, cy2).perpendicular(mx, my, length=1)


  def to_drawables(self, **kwargs):
    return (drawsvg.Line(self.x1, self.y1, self.x2, self.y2, **kwargs),)

  def draw_to_path(self, path, include_m=True, include_l=False):
    if include_l:
      path.L(self.x1, self.y1)
    elif include_m:
      path.M(self.x1, self.y1)
    path.L(self.x2, self.y2)

# end of ESegment


# --------------------------------------------

class ECircle:

  def __init__(self, cx, cy, r, cw=True):
    self.cx = cx
    self.cy = cy
    if r < 0:
      r = -r
      cw = not cw
    self.r = r
    self.cw = cw

  def __str__(self):
    return f'{type(self).__name__}(({self.cx:.3f},{self.cx:.3}), radius={self.r:.3f}, cw={self.cw})'

  def __repr__(self):
    return str(self)


  def reverse(self):
    if self.cw is not None:
      self.cw = not self.cw

  def reversed(self):
    if self.cw is not None:
      cw = not self.cw
    else:
      cw = None
    return ECircle(self.cx, self.cy, self.r, cw=cw)


  def getCoords(self, n=20):
    incr = 2*math.pi/n
    pts = []
    for i in range(n):
      pts.append( ( self.cx + math.cos(i*incr)*self.r,
                    self.cy + math.sin(i*incr)*self.r))
    return pts


  def contains(self, x, y):
    # is a pt in the circle?
    return nearZero((x - self.cx)**2 + \
                    (y - self.cy)**2 - \
                    self.r**2 )

  def onSegment(self, x, y):
    return self.contains(x, y)


  def tangentAt(self, x, y):
    # get tangent line to the circle at this pt.
    if not self.contains(x, y):
      raise ValueError("Point is not on circle")
    if nearZero(y - self.cy):
      return EVertical.at(x)
    slope = -(x - self.cx)/(y - self.cy)
    return ELine(x, y, slope)



  def invert(self, ic):
    # invert the circle to create a line or a circle
    pts1 = self.getCoords(10)
    iPts = []
    for i, pt in enumerate(pts1):
      ipt = invertPt(*pt, ic.cx, ic.cy, ic.r)
      if not ipt == None:
        iPts.append(ipt)
    if orientation(*iPts[0], *iPts[1], *iPts[2]) == 'collinear':
      return ELine.through(*iPts[0], *iPts[1])
    else:
      return ECircle.points(*iPts[0], *iPts[1], *iPts[2])



  def trimmed(self, x1, y1, x2, y2, cw=None, 
                        choose_shorter=None, **kwargs):
    '''Return an arc of the circle, assuming the 
       given points are on it  '''
    assert cw is None or choose_shorter is None
    if cw is None: 
      cw = self.cw
    cx, cy, r = self.cx, self.cy, self.r
    startDeg = math.degrees(math.atan2(y1-cy, x1-cx))
    endDeg = math.degrees(math.atan2(y2-cy, x2-cx))
    if choose_shorter:
      cw = (endDeg - startDeg) % 360 <= 180
    return EArc(cx, cy, r, startDeg, endDeg, cw=cw)


  @staticmethod
  def _centerRadFromPoints(x1, y1, x2, y2, x3, y3):
    # https://planetcalc.com/8116/
    c1 = 2*x1 - 2*x2
    c2 = 2*y1 - 2*y2
    c3 = x2**2 - x1**2 + y2**2 - y1**2
    c4 = 2*x1 - 2*x3
    c5 = 2*y1 - 2*y3
    c6 = x3**2 - x1**2 + y3**2 - y1**2
    cx, cy = numpy.linalg.solve([[c1,c2],[c4,c5]], [-c3,-c6])
    rad = math.hypot(x1-cx, y1-cy)
    return cx, cy, rad


  @classmethod
  def points(cls, x1, y1, x2, y2, x3, y3, **kwargs):
    # Returns circle passing the given 3 points.
    cx, cy, rad = cls._centerRadFromPoints(x1, y1, 
                                      x2, y2, x3, y3)
    return cls(cx, cy, rad, **kwargs)


  def to_drawables(self, **kwargs):
    return (drawsvg.Circle(self.cx, self.cy, self.r, **kwargs),)


  def draw_to_path(self, path, cutDeg=0, 
                   include_m=True, include_l=False):
    path.arc(self.cx, self.cy, self.r, cutDeg, cutDeg+180,
         cw=self.cw, include_m=include_m, include_l=include_l)

    path.arc(self.cx, self.cy, self.r, cutDeg+180, cutDeg,
         cw=self.cw, include_m=False, include_l=False)


# end of ECircle


# --------------------------------------------

class EArc(ECircle):

  def __init__(self, cx, cy, r, startDeg, endDeg, cw=True):
    super().__init__(cx, cy, r)
    self.startDeg = startDeg
    self.endDeg = endDeg
    self.cw = cw

  def __str__(self):
    return f'{type(self).__name__}(({self.cx:.3f},{self.cy:.3f}), radius={self.r:.3f}, {self.startDeg:.1f}--{self.endDeg:.1f},cw={self.cw})'


  def startPt(self):
    x = self.r * math.cos(math.radians(self.startDeg))
    y = self.r * math.sin(math.radians(self.startDeg))
    return self.cx+x, self.cy+y

  def endPt(self):
    x = self.r * math.cos(math.radians(self.endDeg))
    y = self.r * math.sin(math.radians(self.endDeg))
    return self.cx+x, self.cy+y

  def midDegree(self):
    if not self.cw:
      startDeg, endDeg = self.endDeg, self.startDeg
    else:
      startDeg, endDeg = self.startDeg, self.endDeg
    diffDeg = (endDeg - startDeg) % 360
    return (startDeg + diffDeg/2) % 360

  def midpoint(self):
    ang = math.radians(self.midDegree())
    return (self.cx + self.r * math.cos(ang),
            self.cy + self.r * math.sin(ang))

  def reverse(self):
    self.startDeg, self.endDeg = self.endDeg, self.startDeg
    self.cw = not self.cw

  def reversed(self):
    return EArc(self.cx, self.cy, self.r, 
         self.endDeg, self.startDeg, cw=not self.cw)

  def onSegment(self, x, y):
    '''Return True if (x, y) is on this arc, 
       assuming the given point is on it '''
    startDeg, endDeg = self.startDeg % 360, self.endDeg % 360
    if self.cw:
      startDeg, endDeg = endDeg, startDeg
    px = x - self.cx
    py = y - self.cy
    pDeg = math.degrees(math.atan2(py, px)) % 360
    if nearZero(pDeg-startDeg) or nearZero(pDeg-endDeg):
      return True
    if startDeg<endDeg:
      return startDeg<=pDeg<=endDeg
    else:
      return (endDeg<=pDeg) ^ (pDeg<=startDeg)


  @classmethod
  def points(cls, sx, sy, ex, ey, mx, my, 
                            exclude_mid=False, **kwargs):
    cx, cy, rad = cls._centerRadFromPoints(sx, sy, ex, ey, mx, my)
    sx, ex, mx = sx-cx, ex-cx, mx-cx
    sy, ey, my = sy-cy, ey-cy, my-cy
    startDeg = math.degrees(math.atan2(sy, sx))
    endDeg = math.degrees(math.atan2(ey, ex))
    midDeg = math.degrees(math.atan2(my, mx))

    # Clockwise around circle if midDeg is not 
    # between start and end
    cw = (midDeg-startDeg)%360 <= (endDeg-startDeg)%360
    cw = cw ^ exclude_mid
    return cls(cx, cy, rad, startDeg, endDeg, cw=cw, **kwargs)


  @classmethod
  def points_with_center(cls, sx, sy, ex, ey, cx, cy,
                r=None, cw=True, **kwargs):
    sx, sy, ex, ey = sx-cx, sy-cy, ex-cx, ey-cy
    sr = math.hypot(sx, sy)
    er = math.hypot(ex, ey)
    if r is None:
      r = (sr + er) / 2
    assert nearZero(sr-r) and nearZero(er-r), 'Rounding error'
    startDeg = math.degrees(math.atan2(sy, sx))
    endDeg = math.degrees(math.atan2(ey, ex))
    return cls(cx, cy, r, startDeg, endDeg, cw=cw, **kwargs)


  def to_drawables(self, **kwargs):
    return (drawsvg.Arc(self.cx, self.cy, self.r, self.startDeg,
         self.endDeg, cw=self.cw, **kwargs),)


  def draw_to_path(self, path, include_m=True, include_l=False):
    path.arc(self.cx, self.cy, self.r, self.startDeg, self.endDeg,
         cw=self.cw, include_m=include_m, include_l=include_l)

# end of EArc


# --------------------------------------------
# intersection functions

class InfiniteIntersections(Exception): pass

class SingleIntersection(Exception): pass

class NoIntersection(Exception): pass

class InsufficientPrecision(Exception): pass


def circleCircle(circ1, circ2):
  '''Return two points where clockwise around 
     circ1 from (x1,y1) to (x2,y2) is
     the arc of circ1 contained in circ2.
     See "Intersection of two circles", Paul Bourke
          https://paulbourke.net/geometry/circlesphere/
  '''
  swap = circ1.r > circ2.r
  if swap:  # Ensure that circ1 is smaller
    circ1, circ2 = circ2, circ1
  r1, cx1, cy1 = circ1.r, circ1.cx, circ1.cy
  r2, cx2, cy2 = circ2.r, circ2.cx, circ2.cy

  # Center-to-center distance
  dx, dy = cx1 - cx2, cy1 - cy2
  d = math.sqrt(dx**2 + dy**2)

  # Check if circles are the same
  if nearZero(d) and nearZero(r1-r2):
    raise InfiniteIntersections(circ2 if swap else circ1)

  # Check if circles are concentric
  if nearZero(d):
    raise NoIntersection()

  # Check for single intersection
  if nearZero(r1+r2-d) or nearZero(r2-r1-d):
    x = (dx / d) * r2 + cx2
    y = (dy / d) * r2 + cy2
    raise SingleIntersection(x, y)

  # Check if small circle fully contained or separate
  if d+r1 < r2 or r1+r2 < d:
    raise NoIntersection()

  # Calculate two intersections
  t = (r1**2 - d**2 - r2**2) / (-2 * d * r2)  # Law of cosines
  if t > 1: 
    t = 1
  elif t < -1: 
    t = -1
  diffAngle = math.acos(t)
  center_rad = math.atan2(dy, dx)
  ix1 = cx2 + r2 * math.cos(center_rad-diffAngle)
  iy1 = cy2 + r2 * math.sin(center_rad-diffAngle)
  ix2 = cx2 + r2 * math.cos(center_rad+diffAngle)
  iy2 = cy2 + r2 * math.sin(center_rad+diffAngle)

  if swap:
    return ix1, iy1, ix2, iy2
  else:
    return ix2, iy2, ix1, iy1


def circleSeg(circ, seg):
  '''Return intersection points.
     See "Intersection of a Line and a Sphere (or circle)", 
          Paul Bourke
          https://paulbourke.net/geometry/circlesphere/
  '''
  ldx, ldy = seg.x2-seg.x1, seg.y2-seg.y1
  if nearZero(ldx) and nearZero(ldy):
    raise InsufficientPrecision()

  cx, cy, r = circ.cx, circ.cy, circ.r
  x1, y1, x2, y2 = seg.x1-cx, seg.y1-cy, seg.x2-cx, seg.y2-cy

  # Calculate angles and min distance from circle center to line
  lineAngle = math.atan2(ldy, ldx)
  ptAngle = math.atan2(y1, x1)
  ptLen = math.sqrt(x1**2 + y1**2)
  d = ptLen * math.sin(ptAngle - lineAngle)

  isNeg = d < 0
  if isNeg: 
    d = -d

  # Check for single intersection (tangent line)
  if nearZero(d - r):
    ix, iy = -r*math.sin(lineAngle), r*math.cos(lineAngle)
    if isNeg:
      ix, iy = -ix, -iy
    raise SingleIntersection(cx+ix, cy+iy)

  # Check if line is separate from circle
  if d > r:
    raise NoIntersection()

  # Calculate two intersections
  diffAngle = math.acos(d / r)
  ix1, iy1 = -r*math.sin(lineAngle-diffAngle), \
              r*math.cos(lineAngle-diffAngle)
  ix2, iy2 = -r*math.sin(lineAngle+diffAngle), \
              r*math.cos(lineAngle+diffAngle)
  if isNeg:
    ix1, iy1, ix2, iy2 = -ix1, -iy1, -ix2, -iy2
  return cx+ix1, cy+iy1, cx+ix2, cy+iy2


def segCircle(seg, circ):
  '''Return two points on seg where a line from 
     (x1,y1) to (x2,y2) would
     be parallel, not antiparallel, to seg.
  '''
  x1, y1, x2, y2 = circleSeg(circ, seg)
  s = ESegment(x1, y1, x2, y2)
  if s.antiparallelTo(seg):
    s.reverse()
  return s.x1, s.y1, s.x2, s.y2


def segSeg(seg1, seg2):
  # Return one intersection point
  rad1 = math.atan2(seg1.y2-seg1.y1, seg1.x2-seg1.x1)
  rad2 = math.atan2(seg2.y2-seg2.y1, seg2.x2-seg2.x1)

  if nearZero((rad1-rad2) % math.pi):
    angle1 = math.atan2(seg2.y1-seg1.y1, seg2.x1-seg1.x1)
    angle2 = math.atan2(seg2.y2-seg1.y1, seg2.x2-seg1.x1)
    if (nearZero((rad1-angle1) % math.pi) or
        nearZero((rad1-angle2) % math.pi)):
      raise InfiniteIntersections(seg1)
    raise NoIntersection()

  c1 = seg1.x2 - seg1.x1
  c2 = -seg2.x2 + seg2.x1
  c3 = seg2.x1 - seg1.x1
  c4 = seg1.y2 - seg1.y1
  c5 = -seg2.y2 + seg2.y1
  c6 = seg2.y1 - seg1.y1
  a, b = numpy.linalg.solve([[c1,c2],[c4,c5]], [c3,c6])
  ix = seg1.x1 + a * (seg1.x2 - seg1.x1)
  iy = seg1.y1 + a * (seg1.y2 - seg1.y1)
  return ix, iy


# --------------------------------------------
# other utilities

def ptStr(x, y):
  return f"({x:.3f}, {y:.3f})"

def polar2cart(r, degAngle):
    x = r * math.cos(math.radians(degAngle))
    y = r * math.sin(math.radians(degAngle))
    return x, y


def orientation(x1, y1, x2, y2, x3, y3):
  # Compute the orientation of three points in sequence
  value = (x2 - x1)*(y3 - y1) - (x3 - x1)*(y2 - y1)
  if value > EPSILON:
    return 'counterclockwise'
  elif value < -EPSILON:
    return 'clockwise'
  else:
    return 'collinear'

def nearZero(val):
  return -EPSILON <= val <= EPSILON

def invertPt(x, y, cx=0, cy=0, r=1):
  # "Geometry", Brannan et al, 2nd ed., p.266
  sqrNorm = (x - cx)**2 + (y - cy)**2
  if nearZero(math.sqrt(sqrNorm)):
    print(f"Cannot invert point ({x:.3f}, {y:.3f})")
    return None
  k = r**2 / sqrNorm
  return k*(x-cx)+cx, k*(y-cy)+cy



# --------------- drawsvg utils ------------------
# https://github.com/cduck/drawsvg

def createDrawing(winWidth=WIN_WIDTH, renderSize=RENDER_SIZE):
  d = drawsvg.Drawing(winWidth, winWidth, origin='center',
                 context=drawsvg.Context(invert_y=True),
                 stroke_width=0.005, stroke='blue',
                 stroke_opacity=1.0, fill='none' )
  d.set_render_size(renderSize, renderSize)
  d.append(drawsvg.Rectangle(-winWidth/2, -winWidth/2, 
                             winWidth, winWidth, 
                         stroke='white', fill='white'))
  maxVal = 2*winWidth
  return d


def saveDrawing(d, fnm):
  print("Saving drawing to", fnm)
  d.save_png(fnm)
  im = img.imread(fnm)
  h, w, c = im.shape
  setWindow(plt, fnm, 50, 50, w, h)
  plt.imshow(im)
  plt.show()


def dotLabel(d, label, x, y, radius=0.02, size=14, color='black'):
  drawDot(d, x, y, radius, color)
  drawLabel(d, label, x, y+(radius*4), size, color)


def drawDot(d, x, y, radius=0.02, color='black'):
  d.draw(ECircle(x, y, radius), stroke=color, fill=color)


def drawLabel(d, label, x, y, size=14, color='black'):
  d.draw(drawsvg.Text(label, size/100, x, y, 
          stroke='none', fill=color, 
          font_weight='bold', center=True))


def drawAxes(d):
  d.draw(ESegment(-d.width,0, d.width,0), stroke='red', 
                             stroke_dasharray='.1,.1')
  d.draw(ESegment(0, -d.width, 0, d.width), stroke='red', 
                             stroke_dasharray='.1,.1')


# --------------- matplotlib utils ------------------

def setWindow(plt, title, x, y, w, h):
  winSize(plt, w, h)
  plt.axis('off')
  plt.tight_layout(pad=0)
  f = plt.gcf()
  f.patch.set_facecolor('lightgray')
  f.canvas.manager.set_window_title(title)
  backend = matplotlib.get_backend()
  if backend == 'TkAgg':
    f.canvas.manager.window.wm_geometry("+%d+%d" % (x, y))
  elif backend == 'WXAgg':
    f.canvas.manager.window.SetPosition((x, y))
  else:  # This works for QT and GTK
    f.canvas.manager.window.move(x, y)


def winSize(plt, w, h, dpi=100):
  widthInch = w / dpi
  heightInch = h / dpi
  plt.figure(figsize=(widthInch, heightInch), dpi=dpi)

