
# pgeom.py
# Andrew Davison, ad@coe.psu.ac.th, Sept. 2024

'''
  Poincare disk geometry classes:
    Point, Ideal, Line, 
    Triangle, Circle, Transform,
    utilities

  Derived from
    https://github.com/cduck/hyperbolic/tree/master
  and 
    "Geometry"
    David A. Brannan, et al,
    Cambridge University Press; 2nd edition
  and
    https://en.wikipedia.org/wiki/Hyperbolic_geometry
  and
    "Euclidean and Non-Euclidean Geometries"
    Marvin Greenberg
    Freeman; 4th edition
'''

import math,cmath
from numpy import linalg
import drawsvg

from geom import *


class Point:

  def __init__(self, x, y):
    self.x = x
    self.y = y
    # Hyperbolic polar coordinates
    theta = math.atan2(y, x)
    r = math.hypot(x, y)
    if self.isIdeal():
      hr = float('inf')
    else:
      hr = 2 * math.atanh(r)
    self.theta = theta
    self.hr = hr

  def __str__(self):
    return f'{type(self).__name__}({self.x:.3f}, {self.y:.3f})'

  def __repr__(self):
    return str(self)

  def __eq__(self, other):
    return (nearZero(self.x - other.x)
        and nearZero(self.y - other.y))

  def __iter__(self):
    return iter((self.x, self.y))

  def isIdeal(self):
    return nearZero(math.hypot(self.x, self.y) - 1)

  def polar_angle_to(self, p2):
    return p2.theta - self.theta

  def distanceTo(self, p2):
    r1, t1 = self.hr, self.theta
    r2, t2 = p2.hr, p2.theta
    # https://en.wikipedia.org/wiki/Coordinate_systems_for_the_hyperbolic_plane
    d = math.acosh(math.cosh(r1)*math.cosh(r2)
             - math.sinh(r1)*math.sinh(r2)*math.cos(t2-t1))
    return d


  def midpoint(self, p2, frac=0.5):
    d = self.distanceTo(p2)
    p_mid = Point.h_polar(d*frac, 0)
    return Transform.moveTo(self, p2).asPoint(p_mid)


  @staticmethod
  def euclid(x, y):
    # generate a point for (x,y) inside the disc
    r = math.hypot(x, y)
    if nearZero(r - 1.0):
      return Ideal(math.atan2(y, x))
    elif r < 1.0:
      return Point(x, y)
    else:
      raise ValueError(
          'Euclidean coords are outside the unit circle')


  @staticmethod
  def polarEuclid(r, theta=None, deg=None):
    # generate a point for (r,theta) inside the disc
    assert (theta is None) != (deg is None)
    if theta is None: 
      theta = math.radians(deg)
    if nearZero(r - 1.0):
      return Ideal(theta)
    elif r < 1.0:
      return Point(r*math.cos(theta), r*math.sin(theta))
    else:
      raise ValueError(
          'Euclidean coords are outside the unit circle')


  @staticmethod
  def h_polar(hr, theta=None, deg=None):
    # generate a point for (hr,theta) using a hyberbolic radius
    assert (theta is None) != (deg is None)
    if theta is None: 
      theta = math.radians(deg)
    r = math.tanh(hr/2)
    x, y = r*math.cos(theta), r*math.sin(theta)
    return Point(x, y)


  def to_drawables(self, radius=0.01, **kwargs):
    return ECircle(self.x, self.y, r=radius).to_drawables(**kwargs)

# end of Point

# ------------------------------------------------

class Ideal(Point):
  # https://en.wikipedia.org/wiki/Ideal_point

  def __init__(self, theta):
    self.theta = theta % (2*math.pi)

  @property
  def x(self):
    return math.cos(self.theta)

  @property
  def y(self):
    return math.sin(self.theta)

  def isIdeal(self):
    return True

  @classmethod
  def degree(cls, deg):
    return cls(math.radians(deg))

  @classmethod
  def radian(cls, rad):
    return cls(rad)

# end of Ideal


# --------------------------------------------

class Line:

  def __init__(self, pShape, segment=False):
    assert isinstance(pShape, (ECircle, EArc, ESegment))
    if segment and not isinstance(pShape, (EArc, ESegment)):
      raise ValueError(
          'Not enough info to determine segment endpoints')

    unitCircle = ECircle(0,0,1)
    if segment:
      if isinstance(pShape, EArc):
        # TODO: Trim if arc extends outside unitCircle circle
        pass
      elif isinstance(pShape, ESegment):
        # TODO: Trim if arc extends outside unitCircle circle
        pass
      else:
        raise ValueError('Unknown type')
    else:
      if isinstance(pShape, ECircle):
        x1, y1, x2, y2 = circleCircle(pShape, unitCircle)
        if not pShape.cw:
          x1, y1, x2, y2 = x2, y2, x1, y1
        pShape = EArc.points_with_center(
                    x1, y1, x2, y2,
                    pShape.cx, pShape.cy, r=pShape.r,
                    cw=pShape.cw)
      elif isinstance(pShape, ESegment):
        x1, y1, x2, y2 = circleSeg(unitCircle, pShape)
        # Keep direction of line the same 
        angle = math.atan2( pShape.y2-pShape.y1,
                               pShape.x2-pShape.x1)
        nAngle = math.atan2(y2-y1, x2-x1)
        ang_diff = (nAngle - angle) % (math.pi*2)
        if math.pi/2 < ang_diff < 3*math.pi/2:
          x1, y1, x2, y2 = x2, y2, x1, y1
        pShape = ESegment(x1, y1, x2, y2)
      else:
        raise ValueError('Unknown type')

    self.pShape = pShape
    self.segment = segment


  def __str__(self):
    return f"Line({self.startPt()}--{self.endPt()}, seg={self.segment})"

  def __repr__(self):
    return str(self)


  @classmethod
  def points(cls, p1, p2, segment=False, **kwargs):
    x1 = p1.x; y1 = p1.y
    x2 = p2.x; y2 = p2.y
    if nearZero(x1-x2) and nearZero(y1-y2):
      raise ValueError('Start and end points are the same')

    if cls._pointsInlineWithOrigin(x1, y1, x2, y2):
      if segment:
        shape = ESegment(x1, y1, x2, y2)
        return cls(shape, segment=True, **kwargs)
      else:
        lineAngle = math.atan2(y2-y1, x2-x1)
        sx, sy = math.cos(lineAngle), math.sin(lineAngle)
        shape = ESegment(-sx, -sy, sx, sy)
        return cls(shape, segment=False, **kwargs)

    r1 = math.hypot(x1, y1)
    r2 = math.hypot(x2, y2)
    if nearZero(r1-1) and nearZero(r2-1):
      # Both points are ideal so invertPt()
      # will not give a 3rd point
      a1 = math.atan2(y1, x1)
      a2 = math.atan2(y2, x2)
      swap = (a2 - a1) % (math.pi*2) > math.pi
      if swap:
        a1, a2 = a2, a1
      a_diff = (a2 - a1) % (math.pi*2)
      assert a_diff <= math.pi
      center_dist = 1/math.cos(a_diff/2)
      center_ang = a1 + a_diff/2
      cx = math.cos(center_ang) * center_dist
      cy = math.sin(center_ang) * center_dist
      r = center_dist * math.sin(a_diff/2)
      arcDeg = 180 - math.degrees(a_diff)
      arc_startDeg = math.degrees(a1) - 90
      arc_endDeg = arc_startDeg - arcDeg

      if swap:
        arc_startDeg, arc_endDeg = arc_endDeg, arc_startDeg
      arc = EArc(cx, cy, r, arc_endDeg, arc_startDeg, cw=not swap)
      return cls(arc, segment=False, **kwargs)
      # Even if segment is True, the given points span the
      # entire line so it isn't a segment
    else:
      ex_mid = True
      if nearZero(r1-1):
        x3, y3 = invertPt(x2, y2)
        if r2 > 1:
          ex_mid = False
      else:
        x3, y3 = invertPt(x1, y1)
        if r1 > 1:
          ex_mid = False

      arc = EArc.points(x1, y1, x2, y2, x3, y3, 
                        exclude_mid=ex_mid)
      if segment:
        return cls(arc, segment=True, **kwargs)
      else:
        x1, y1, x2, y2 = circleCircle(arc, ECircle(0,0,1))
        if not arc.cw:
          x1, y1, x2, y2 = x2, y2, x1, y1
        arc_full = EArc.points(
            x1, y1, x2, y2, x3, y3, exclude_mid=ex_mid)
        assert arc.cw == arc_full.cw
        return cls(arc_full, segment=False, **kwargs)


  def startPt(self):
    return Point(*self.pShape.startPt())


  def endPt(self):
    return Point(*self.pShape.endPt())


  def midpoint(self):
    x, y = self.pShape.midpoint()
    return Point(x, y)

  def length(self):
    p1 = self.startPt()
    p2 = self.endPt()
    return p1.distanceTo(p2)


  def intersectPts(self, dline):
    '''Return a list of intersection points.
       Assumes the d-line is infinite length, not segments.
    '''
    x2, y2 = 1, 1  # Default, invalid
    try:
      if isinstance(self.pShape, ECircle):
        if isinstance(dline.pShape, ECircle):
          x1, y1, x2, y2 = circleCircle(self.pShape, dline.pShape)
        else:
          x1, y1, x2, y2 = circleSeg(self.pShape, dline.pShape)
      else:
        if isinstance(dline.pShape, ECircle):
          x1, y1, x2, y2 = circleSeg(dline.pShape, self.pShape)
        else:
          x1, y1 = segSeg(self.pShape, dline.pShape)
    except InfiniteIntersections as e:
      raise e from None
    except SingleIntersection as e:
      x1, y1 = e.args
    except NoIntersection:
      x1, y1 = 1, 1  # Invalid

    pts = []
    try:
      pts.append(Point.euclid(x1, y1))
    except ValueError: pass
    try:
      pts.append(Point.euclid(x2, y2))
    except ValueError: pass
    return pts


  def onSegment(self, p):
    return self.pShape.onSegment(p.x, p.y)


  def segIntersects(self, dline):
    '''Return the list of intersections points
       Only returns points actually on the segments, 
       not their infinite extension. '''
    valid = [ p for p in self.intersectPts(dline)
                if (self.onSegment(p) and dline.onSegment(p)) ]
    return valid


  def trimmed(self, p1, p2, **kwargs):
    '''Return a segment going 
       from Poincare disk coordinates x1,y1 to x2,y2.
       Assumes x1,y1 and x2,y2 are on the dline. '''
    x1 = p1.x; y1 = p1.y
    x2 = p2.x; y2 = p2.y
    tShape = self.pShape.trimmed(x1, y1, x2, y2, **kwargs)
    return type(self)(tShape, segment=True)


  def reverse(self):
    self.pShape.reverse()

  def reversed(self):
    new_shape = self.pShape.reversed()
    new_obj = type(self)(new_shape, segment=True)
    new_obj.segment = self.segment
    return new_obj


  def perpendicular(self, pt):
    x = pt.x; y = pt.y
    if nearZero(x) and nearZero(y):
      s1 = None
    else:
      s1 = ESegment.radicalAxis(ECircle(0,0,1), (x,y))
    s2 = ESegment.radicalAxis(self.pShape, (x,y))

    if s1 is None or s1.parallelTo(s2):
      # Infinite radius circle
      shape = s2.perpendicular(x,y)

      # Ensure correct direction
      if isinstance(self.pShape, ESegment):
        if shape.antiparallelTo(self.pShape):
          shape.reverse()
      else:  # pShape is an ECircle
        radial = ESegment.points(
            x, y, self.pShape.cx, self.pShape.cy)
        if radial.antiparallelTo(shape) ^ (not self.pShape.cw):
          shape.reverse()
      return Line(shape, segment=False)
    else:
      cx, cy = segSeg(s1, s2)
      r = math.hypot(cy-y, cx-x)
      shape = ECircle(cx, cy, r)

      # Ensure correct direction
      shape.cw = True
      if isinstance(self.pShape, ESegment):
        x1, y1, x2, y2 = segCircle(
                             self.pShape, shape)
      else:  # pShape is an ECircle
        x1, y1, x2, y2 = circleCircle(
                             self.pShape, shape)
        shape.cw = shape.cw ^ (not self.pShape.cw)

      if math.hypot(x1, y1) > math.hypot(x2, y2):
        shape.cw = not shape.cw
      return Line(shape, segment=False)


  @staticmethod
  def _pointsInlineWithOrigin(x1, y1, x2, y2):
    if ((nearZero(x1) and nearZero(x2)) or
        (nearZero(y1) and nearZero(y2)) or
        (nearZero(x1) and nearZero(y1)) or
        (nearZero(x2) and nearZero(y2))):
      return True
    elif ((nearZero(x1) or nearZero(x2)) and
          (nearZero(y1) or nearZero(y2))):
      return False
    elif nearZero(x1) or nearZero(x2):
      return nearZero(x1/y1 - x2/y2)
    else:
      return nearZero(y1/x1 - y2/x2)


  def to_drawables(self, **kwargs):
    shape = self.pShape
    return shape.to_drawables(**kwargs)


  def draw_to_path(self, path, **kwargs):
    shape = self.pShape
    return shape.draw_to_path(path, **kwargs)


# end of Line

# --------------------------------------------

class Triangle:
  # https://en.wikipedia.org/wiki/Hyperbolic_triangle

  def __init__(self, pt0, pt1, pt2):
    verts = [pt0, pt1, pt2]
    edges = [None, None, None]
    for i in range(3):
      startP, endP = verts[i], verts[(i+1)%3]
      edges[i] = Line.points(startP, endP, segment=True)
      if not isinstance(edges[i], Point):
        edges[i] = edges[i].trimmed(startP, endP, choose_shorter=True)

    self.edges = tuple(edges)
    self.vertices = tuple(verts)


  def __str__(self):
    return "Tri(" + str(self.vertices) + ")"

  def __repr__(self):
    return str(self)


  def getVertices(self):
    return list(self.vertices)

  def getVertex(self, idx):
    if (idx < 0) or (idx >= 3):
      raise ValueError('Vertex index " + str(idx) + " out of range')
    else:
      return self.vertices[idx]

  def getEdges(self):
    return list(self.edges)

  def getEdge(self, idx):
    if (idx < 0) or (idx >= 3):
      raise ValueError('Edges index " + str(idx) + " out of range')
    else:
      return self.edges[idx]


  def getAngle(self, idx):
    if (idx < 0) or (idx >= 3):
      raise ValueError('Vertex index " + str(idx) + " out of range')
    else:
      pL = self.vertices[(idx-1)%3]
      p = self.vertices[idx]
      pR = self.vertices[(idx+1)%3]
      a1 = p.polar_angle_to(pL)
      a2 = p.polar_angle_to(pR)
      # https://byjus.com/maths/angle-between-two-lines/
      # print("polars", math.degrees(a1), math.degrees(a2))
      angle = abs( math.atan2(math.tan(a2)-math.tan(a1), 
                        1-math.tan(a1)*math.tan(a2)))
      return angle
  

  def area(self):
    return math.pi - \
      (self.getAngle(0) + self.getAngle(1) + self.getAngle(2))

  def transform(self, trans):
    verts = trans.asList(self.getVertices())
    return Triangle(*verts)


  def to_drawables(self, **kwargs):
    path = drawsvg.Path(**kwargs)
    self.draw_to_path(path, **kwargs)
    path.Z()
    return (path,)


  def draw_to_path(self, path, include_m=True, include_l=False,
           **kwargs):
    for i, edge in enumerate(self.edges):
      if isinstance(edge, Point):
        continue
      edge.draw_to_path(path, 
          include_m=include_m and i==0,
          include_l=include_l and i==0)

# end of Triangle

# --------------------------------------------

class Circle:

  def __init__(self, pShape, center=None, r=None):
    if not isinstance(pShape, ECircle):
      raise ValueError('Shape must be a euclidean circle')
    self.pShape = pShape
    if r is None or center is None:
      de0 = math.hypot(pShape.cx, pShape.cy)
      de1 = de0 - pShape.r
      de2 = de0 + pShape.r
      dh1 = radial_euclid_to_poincare(de1)
      dh2 = radial_euclid_to_poincare(de2)
    if r is None:
      r = (dh2 - dh1) / 2
    if center is None:
      cr = (dh2 + dh1) / 2
      theta = math.atan2(pShape.cy, pShape.cx)
      center = Point.h_polar(cr, theta)
    self.r = r
    self.center = center


  @staticmethod
  def center_radius(pt, radius, cw=True):
    '''Create a hyperbolic circle from a hyperbolic point 
       and length within the hyperbolic plane. '''
    if radius < 0:
      radius = -radius
      cw = not cw
    cr = pt.hr
    dh1 = cr - radius
    dh2 = cr + radius
    de1 = radial_poincare_to_euclid(dh1)
    de2 = radial_poincare_to_euclid(dh2)
    er = (de2 - de1) / 2
    ecr = (de2 + de1) / 2
    x = ecr * math.cos(pt.theta)
    y = ecr * math.sin(pt.theta)
    shape = ECircle(x, y, er, cw=cw)
    return Circle(shape, center=pt, r=radius)


  # def to_drawables(self, transform=None, **kwargs):
  def to_drawables(self, **kwargs):
    shape = self.pShape
    # if transform:
      # shape = transform.asShape(shape)
    return shape.to_drawables(**kwargs)

# end of Circle

# --------------------------------------------

class Transform:
  # https://en.wikipedia.org/wiki/M%C3%B6bius_transformation

  ''' Mobius transformation as a matrix 
      "Geometry", Brannan et al, 2nd ed., p.298, p.300
       f(z) = (az + b)/(cz + d)  =  [a b] [z]
                                    [c d] [1]

       ad - bc /= 0
    - conformal (i.e. preserves angles) and is a group so
      transformations can be composed

    Basic ops:
      1. z --> cz, c in R, scaling
      2. z --> z + A, A in C, translation
      3. z --> Az, A = e^i*theta, rotation
      4. z --> z.conj, complex conjugation
      5. z --> 1/z , inversion
      6. mirror z in a line segment
  ''' 

  def __init__(self, a, b, c, d, e=1, conj=False):
    self.abcd = (a*e, b*e, c, d)  # e is rotation factor
    self.conj = conj

  def __repr__(self):
    return '{}({}, {}, {}, {}{})'.format(type(self).__name__,
        # show abcd tuple values as complex numbers
        *map(lambda z:complex(round(z.real,3),
                              round(z.imag,3)), self.abcd),
        ', conj' if self.conj else '')


  def asTuple(self, point):
    # apply complex transform to a point
    z = complex(*point)
    if self.conj:
      z = z.conjugate()
    a,b,c,d = self.abcd
    numer = a*z + b
    denom = c*z + d
    if nearZero(abs(denom)):
      zt = numer * 1e5
    else:
      zt = numer / denom
    return zt.real, zt.imag


  def asPoint(self, point):
    return Point(*self.asTuple(point))


  def asIdeal(self, point, verify=False):
    x, y = self.asTuple(point)
    if verify and not nearZero(math.hypot(x, y) - 1):
      raise ValueError('Invalid transform')
    return Ideal(math.atan2(y, x))


  def asList(self, points, verify=False):
    if len(points) > 0:
      if isinstance(points[0], Ideal):
        if verify:
          def f(p): return self.asIdeal(p, verify=True)
        else:
          f = self.asIdeal
      elif isinstance(points[0], Point):
        f = self.asPoint
      else:
        f = self.asTuple
    return [f(p) for p in points]


  def asShape(self, shape):
    '''Transform a euclidean shape.
       Supports euclidean Circle, EArc, and Line. '''
    if isinstance(shape, (EArc, ESegment)):
      closed = False
      p1 = shape.startPt()
      p2 = shape.midpoint()
      p3 = shape.endPt()
    elif isinstance(shape, ECircle):
      closed = True
      p1 = shape.cx+shape.r, shape.cy
      p2 = shape.cx, shape.cy+shape.r
      p3 = shape.cx-shape.r, shape.cy
      if not shape.cw:
        p1, p3 = p3, p1
    else:
      raise TypeError('Unsupported shape: {}'.format(shape))

    # Transform
    p1, p2, p3 = self.asList((p1, p2, p3))

    # create shape
    try:  # See if it is an arc or circle
      out = EArc.points(*p1, *p3, *p2)
      if closed:  # It's a circle
        out = ECircle(out.cx, out.cy, out.r, cw=out.cw)
      return out
    except (InfiniteIntersections, linalg.LinAlgError):
      # It is a line
      return ESegment(*p1, *p3)


  def __call__(self, *points, verify=False):
    # we can call obj(a, b, c) instead of obj.__call__(a, b, c)
    return self.asList(points, verify=verify)


  def inverted(self):
    ''' Inverse matrix (p.300):  [ d -b]
                                 [-c  a]
    '''
    a,b,c,d = self.abcd
    if self.conj:
      a,b,c,d = a.conjugate(), b.conjugate(), \
                c.conjugate(), d.conjugate()
    return Transform(d, -b, -c, a, conj=self.conj)


  def conjugate(self):
    a, b, c, d = self.abcd
    return Transform(a.conjugate(), b.conjugate(),
             c.conjugate(), d.conjugate(), conj=not self.conj)

  @staticmethod
  def identity():
    return Transform(1,0,0,1,1)

  @staticmethod
  def compose(*ts):
    # compose in left-to-right order so rightmost op is done first
    a,b,c,d = 1,0,0,1
    conj = False
    for i, trans in enumerate(ts):
      a2,b2,c2,d2 = trans.abcd
      if trans.conj:
        a,b,c,d = (a.conjugate(), b.conjugate(), 
                   c.conjugate(), d.conjugate())
        conj = not conj
      a,b,c,d = a*a2+c*b2, b*a2+d*b2, a*c2+c*d2, b*c2+d*d2
      '''  [a b] = [a2 b2] [a b] 
           [c d] = [c2 d2] [c d] 
      '''
    return Transform(a,b,c,d,conj=conj)

  @staticmethod
  def originStep(nOrigin):
    ''' [1      -k]    maps 0 to -k and k to 0
        [-k.conj 1]    'from k'
    '''
    k = complex(*nOrigin)
    a,b,c,d = 1, -k, (-k).conjugate(), 1
    e = 1
    return Transform(a,b,c,d,e)

  @classmethod
  def moveTo(cls, offset):
    return cls.originStep(offset).inverted()

  @staticmethod
  def rotate(rad=None, deg=None):
    assert (rad is None) + (deg is None) == 1
    if deg is not None:
      rad = math.radians(deg)
    e = cmath.rect(1, rad)  
    # polar --> rectangular (gets e^*(i*rad))
    return Transform(1,0,0,1,e)

  @staticmethod
  def mirror(p, q):
    # mirror in line p--q
    # a version of the reflection lemma on p.379
    zp, zq = complex(*p), complex(*q)
    zpc, zqc = zp.conjugate(), zq.conjugate()
    a = (zp*zq)*zpc - (zp*zq)*zqc - zp + zq
    b = zp*zqc - zq*zpc
    return Transform(a, b, -b, a.conjugate(), conj=True)

# end of Transform


# --------------------------------------------
# utilities

def radial_euclid_to_poincare(r):
  return 2 * math.atanh(r)  # p.369

def radial_poincare_to_euclid(r):
  return math.tanh(r/2)


def angleSidesTri(a, b, c):
  ''' Return angle C using 3 hyperbolic triangle 
      lengths a, b, c
      https://en.wikipedia.org/wiki/Hyperbolic_triangle
  '''
  C = math.acos(((math.cosh(a)*math.cosh(b))-math.cosh(c))/
                 (math.sinh(a)*math.sinh(b)) )
    # Theorem 10.4, ch.10, Greenberg
  return C


def sideTri(adj1, adj2, opposite):
  '''Solve for a hyperbolic triangle side length 
     from three angles.
     Uses the hyperbolic law of cosines to solve for the 
     side length.  Input angles in radians.
  '''
  A, B, C = adj1, adj2, opposite
  c = math.acosh((math.cos(C) + math.cos(A)*math.cos(B)) /
           (math.sin(A)*math.sin(B)))
    # Theorem 10.4, ch.10, Greenberg
  return c


def angleTri(adj1, adj2, opposite_len):
  '''Solve for a hyperbolic triangle angle from the other 
     two angles and opposite side length.
     Uses the hyperbolic law of cosines to solve for the 
     angle opposite the given side.  Input angles in radians.
  '''
  c, A, B = opposite_len, adj1, adj2
  C = math.acos(math.cosh(c)*math.sin(A)*math.sin(B) -
          math.cos(A)*math.cos(B))
    # Theorem 10.4, ch.10, Greenberg, 4th ed.
  return C


def anglesArea(A, B, C):
  return math.pi - (A + B + C)



def polyVerts(p, q):
  r = polyRadius(p, q)
  return [ Point.h_polar(r, deg=-i*360/p)
                                for i in range(p) ]

def polyRadius(p, q):
  ''' Calculate the hyperbolic circumcircle radius
    for a regular polygon by considering the right-angled
    triangle that makes up half of one side. The angles
    of that triangle are A, B, and C.

    p = the number of sides of the polygon;
    q = the number of polygons that meet at
        a vertex

    For a hyperbolic regular polygon
    (p-2) * (q-2) > 4
  '''
  # returns hyperbolic circumcircle radius
  A = math.pi/p
  B = math.pi/q
  C = math.pi/2
  r = math.acosh( (math.cos(A)*math.cos(B) + math.cos(C))/  \
                    (math.sin(A)*math.sin(B)) )
      # Theorem 10.4, Ch 10, Greenberg
  # print("r", r)
  return r


def toEdges(pts):
  n = len(pts)
  return [Line.points(pts[i], pts[(i+1)%n], segment=True) 
                                            for i in range(n)]
