
# tree.py
# Andrew Davison, ad@coe.psu.ac.th, Sept. 2024
'''
  Based on the in[11] and in[12] code at
  https://github.com/cduck/hyperbolic/blob/master/examples/poincare.ipynb
  that generates tree.svg

  Generates tree.png
'''

import math
from geom import *
from pgeom import *


class TreeNode:
  # a tree node is a point, a list of children tree nodes,
  #  and its parent node

  def __init__(self, point, children=None, parent=None):
    self.point = point
    self.children = [] if children is None else children
    self.parent = parent

  def transformed(self, trans, nParent):
    # apply the transformation, making a new node
    nPt = trans.asPoint(self.point)
    nNode = TreeNode(nPt, parent=nParent)
    nNode.children = [child.transformed(trans, nNode) 
                             for child in self.children]
    return nNode


  def to_drawables(self, **kwargs):
    # build a list of drawable tree edges 
    ds = []
    for child in self.children:
      edge = Line.points(self.point, child.point, 
                             segment=True)
      ds0 = edge.to_drawables(**kwargs)
      ds.extend(ds0)
    for child in self.children:
      ds.extend(child.to_drawables(**kwargs))
    return ds

# end of TreeNode
# ------------------------------------------

def genTree(root, depth):
  # build binary tree bottom-up
  pt = root.point
  for d in range(depth):
    nRoot = TreeNode(pt)
    childUp = root.transformed(moveUp, nRoot)
    childDown = root.transformed(moveDown, nRoot)
    nRoot.children = [childDown, childUp]
    root = nRoot
  return root


depth = 4  # number of sub-branches in binary tree
innerAngle = 30  # angle to x-axis 
numRepeats = 360//innerAngle

d = createDrawing()
d.draw(ECircle(0, 0, 1), stroke='silver') # disc boundary


origPt = Point(0, 0)
p1 = Point.h_polar(0.8, deg=0)  
      # point at end of origPt branch

rotB = Transform.rotate(deg=180+innerAngle/2)  
       # rotate below - x-axis
rotA = Transform.rotate(deg=180-innerAngle/2)
       # rotate above - x-axis
# print(rotB, rotA)

# rotated versions of p1
ptB, = rotB(p1)   
ptA, = rotA(p1)
# d.draw(ptB)
# d.draw(ptA)

moveUp = Transform.originStep(ptB)
   # shift the below point to get a move upwards
moveDown = Transform.originStep(ptA)
   # shift the above point to get a move downwards

# generate a single binary tree starting at p1
binTree = genTree(TreeNode(p1), depth)

# use binTree to create numRepeats branches connected to origPt
tree = TreeNode(origPt, [], None)
for i in range(numRepeats):
  rot = Transform.rotate(deg = -innerAngle*i)
  treeRot = binTree.transformed(rot, tree)
  tree.children.append(treeRot)  # add rotated binary tree
d.draw(tree)

saveDrawing(d, 'tree.png')
